package com.example.myfirstapp;

import android.app.AlertDialog;
import android.app.Application;
import android.app.DownloadManager;
import android.content.Context;
import android.content.DialogInterface;
import android.hardware.Sensor;
import android.hardware.SensorEvent;
import android.hardware.SensorEventListener;
import android.hardware.SensorManager;
import android.hardware.usb.UsbDeviceConnection;
import android.hardware.usb.UsbManager;
import android.location.Location;
import android.os.AsyncTask;
import android.os.Bundle;
import android.os.Environment;
import android.support.design.widget.TabLayout;
import android.support.v4.view.ViewPager;
import android.support.v7.app.AppCompatActivity;
import android.content.pm.ActivityInfo;
import android.util.Log;
import android.view.View;
import android.widget.ScrollView;
import android.widget.TextView;
import android.widget.Toast;

import com.google.android.gms.common.ConnectionResult;
import com.google.android.gms.common.api.GoogleApiClient;
import com.google.android.gms.common.api.Response;
import com.google.android.gms.maps.CameraUpdate;
import com.google.android.gms.maps.CameraUpdateFactory;
import com.google.android.gms.maps.GoogleMap;
import com.google.android.gms.maps.OnMapReadyCallback;
import com.google.android.gms.maps.SupportMapFragment;
import com.google.android.gms.maps.model.BitmapDescriptorFactory;
import com.google.android.gms.maps.model.LatLng;
import com.google.android.gms.maps.model.LatLngBounds;
import com.google.android.gms.maps.model.Marker;
import com.google.android.gms.maps.model.MarkerOptions;

import com.hoho.android.usbserial.driver.UsbSerialDriver;
import com.hoho.android.usbserial.driver.UsbSerialPort;
import com.hoho.android.usbserial.driver.UsbSerialProber;
import com.hoho.android.usbserial.util.SerialInputOutputManager;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import com.google.android.gms.location.LocationListener;
import com.google.android.gms.location.LocationRequest;
import com.google.android.gms.location.LocationServices;

import org.json.*;

import static java.lang.Math.atan2;
import static java.lang.Math.cos;
import static java.lang.Math.sin;
import static java.lang.Math.sqrt;

public class MainActivity extends AppCompatActivity implements
        OnMapReadyCallback,
        GoogleApiClient.OnConnectionFailedListener,
        GoogleApiClient.ConnectionCallbacks,
        LocationListener,
        SensorEventListener {

    private static final String TAG = "MyActivity";

    protected Application mApplication;

    MyPagerAdapter adapter;

    GoogleMap map;

    private final ExecutorService mExecutor = Executors.newSingleThreadExecutor();
    private static UsbSerialPort sPort = null;
    protected SerialPort mSerialPort;
    protected OutputStream mOutputStream;
    private InputStream mInputStream;

    ScrollView mScrollView;
    TextView mConsoleText;
    String logContent = "";
    String jsonStr = "";

    Marker selfMarker;
    Marker balloonMarker;
    double selfAltitude = 0;

    GoogleApiClient mGoogleApiClient;

    FileOutputStream fileOutputStream = null;
    File publicStorageDirectory = null;

    private SensorManager mSensorManager;
    private final float[] mAccelerometerReading = new float[3];
    private final float[] mMagnetometerReading = new float[3];

    private final float[] mRotationMatrix = new float[9];
    private final float[] mOrientationAngles = new float[3];
    private float compassDirection = 0;

    private long lastNewDataTime = 0;
    private int highestGSMReportId = 0;
    private long lastFileWriteTime = 0;

    private SerialInputOutputManager mSerialIoManager;
    private final SerialInputOutputManager.Listener mListener =
            new SerialInputOutputManager.Listener() {
                @Override
                public void onRunError(Exception e) {
                    Log.d(TAG, "Runner stopped.");
                }
                @Override
                public void onNewData(final byte[] data) {
                    MainActivity.this.runOnUiThread(new Runnable() {
                        @Override
                        public void run() {
                            MainActivity.this.handleReceivedData(data);
                        }
                    });
                }
            };

    private void toast(String text){
        Toast toast = Toast.makeText(this, text, Toast.LENGTH_LONG);
        toast.show();
    }

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);

        mApplication = (Application) getApplication();

        buildGoogleApiClient();

        setRequestedOrientation(ActivityInfo.SCREEN_ORIENTATION_PORTRAIT);

        setContentView(R.layout.activity_main);

        SupportMapFragment mapFragment = (SupportMapFragment)getSupportFragmentManager().findFragmentById(R.id.map);
        mapFragment.getMapAsync(this);

        TabLayout tabLayout = (TabLayout) findViewById(R.id.tab_layout);
        tabLayout.addTab(tabLayout.newTab().setText("Location"));
        tabLayout.addTab(tabLayout.newTab().setText("Sensors"));
        tabLayout.addTab(tabLayout.newTab().setText("Tracking"));
        tabLayout.addTab(tabLayout.newTab().setText("Log"));
        tabLayout.setTabGravity(TabLayout.GRAVITY_FILL);

        final ViewPager viewPager = (ViewPager) findViewById(R.id.pager);
        adapter = new MyPagerAdapter(getSupportFragmentManager(), tabLayout.getTabCount());
        viewPager.setAdapter(adapter);
        viewPager.addOnPageChangeListener(new TabLayout.TabLayoutOnPageChangeListener(tabLayout));
        tabLayout.addOnTabSelectedListener(new TabLayout.OnTabSelectedListener() {
            @Override
            public void onTabSelected(TabLayout.Tab tab) {
                viewPager.setCurrentItem(tab.getPosition());
                //logConsole("Moved to tab "+tab.getPosition());
                refreshData();

                if ( tab.getPosition() == 2 ) {
                    TabFragment3 tab3 = (TabFragment3) adapter.tab3;

                    tab3.mapTypeButton.setOnClickListener( new View.OnClickListener() {
                        public void onClick(View v) {
                            // Code here executes on main thread after user presses button
                            showMapTypeSelectorDialog();
                        }
                    });

                    tab3.followTypeButton.setOnClickListener( new View.OnClickListener() {
                        public void onClick(View v) {
                            // Code here executes on main thread after user presses button
                            showFollowTypeSelectorDialog();
                        }
                    });
                }
            }

            @Override
            public void onTabUnselected(TabLayout.Tab tab) {

            }

            @Override
            public void onTabReselected(TabLayout.Tab tab) {

            }
        });

        findUsbSerial();

        mSensorManager = (SensorManager) getSystemService(Context.SENSOR_SERVICE);

        new CheckServerReports().execute("http://www.iforce2d.net/bt/latest.php");

        boolean externalWritable = isExternalStorageWritable();
        boolean externalReadable = isExternalStorageReadable();
        Log.d(TAG, "External readable: "+externalReadable);
        Log.d(TAG, "External writable: "+externalWritable);

        publicStorageDirectory = getPublicStorageDir();

        Date d = new Date();

        String strDateFormatFilename = "yyyyMMdd_HHmmss";
        DateFormat dateFormatFilename = new SimpleDateFormat(strDateFormatFilename);
        String logfileName = "balloonLog_"+dateFormatFilename.format(d)+".txt";

        File file = new File(publicStorageDirectory, logfileName);

        try {
            fileOutputStream = new FileOutputStream(file);
            String strDateFormat = "yyyy/MM/dd HH:mm:ss";
            DateFormat dateFormat = new SimpleDateFormat(strDateFormat);
            String s = dateFormat.format(d);
            fileOutputStream.write(("Log started at "+s+"\n").getBytes());
            fileOutputStream.flush();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /* Checks if external storage is available for read and write */
    public boolean isExternalStorageWritable() {
        String state = Environment.getExternalStorageState();
        if (Environment.MEDIA_MOUNTED.equals(state)) {
            return true;
        }
        return false;
    }

    /* Checks if external storage is available to at least read */
    public boolean isExternalStorageReadable() {
        String state = Environment.getExternalStorageState();
        if (Environment.MEDIA_MOUNTED.equals(state) ||
                Environment.MEDIA_MOUNTED_READ_ONLY.equals(state)) {
            return true;
        }
        return false;
    }

    public File getPublicStorageDir() {
        // Get the directory for the user's public pictures directory.
        File file = new File( Environment.getExternalStoragePublicDirectory(Environment.DIRECTORY_DOCUMENTS), "balloon");
        if (!file.mkdirs()) {
            //Log.e(TAG, "Directory not created");
        }
        Log.d(TAG, "Public storage directory: "+file.getAbsolutePath());
        return file;
    }

    @Override
    public void onMapReady(GoogleMap googleMap) {
        map = googleMap;
        map.setMapType(GoogleMap.MAP_TYPE_HYBRID);
        LatLng hamilton = new LatLng(-37.771869, 175.270158);
        MarkerOptions selfMarkerOptions = new MarkerOptions();
        MarkerOptions balloonMarkerOptions = new MarkerOptions();
        balloonMarkerOptions.icon( BitmapDescriptorFactory.defaultMarker(BitmapDescriptorFactory.HUE_AZURE) );
        balloonMarker = map.addMarker(balloonMarkerOptions.position(hamilton).title("Balloon"));
        selfMarker =    map.addMarker(selfMarkerOptions.position(hamilton).title("Me"));
        map.moveCamera(CameraUpdateFactory.newLatLngZoom(hamilton, 8));
    }

    private void handleReceivedData(byte[] data) {

        jsonStr += new String(data);

        int closeBracketPos = jsonStr.indexOf("}");
        if ( closeBracketPos > -1 ) {
            String tmp = jsonStr.substring(0, closeBracketPos+1);
            jsonStr = jsonStr.substring(closeBracketPos+1);

            parseData(tmp);
        }

        logConsole( new String(data) );

    }

    public static int countLines(String str) {
        if(str == null || str.isEmpty())
        {
            return 0;
        }
        int lines = 1;
        int pos = 0;
        while ((pos = str.indexOf("\n", pos) + 1) != 0) {
            lines++;
        }
        return lines;
    }

    public void logConsole(String s) {

        //Log.d(TAG, s);

        logContent += "\n" + s;

        // Erase excessive lines
        int numLines = countLines(logContent);
        while ( numLines > 30 ) {
            int nlPos = logContent.indexOf("\n");
            logContent = logContent.substring(nlPos + 1);
            numLines--;
        }

        if ( adapter != null && adapter.tab4 != null ) {
            //Log.d(TAG, "Checking adapter.tab4");
            mScrollView = ((TabFragment4) adapter.tab4).consoleScrollView;
            mConsoleText = ((TabFragment4) adapter.tab4).consoleTextView;
        }

        if ( mConsoleText == null )
            return;

        mConsoleText.setText(logContent);
        //mScrollView.fullScroll(View.FOCUS_DOWN);
    }

    public void findUsbSerial() {

        logConsole("findUsbSerial");

        // Find all available drivers from attached devices.
        UsbManager manager = (UsbManager) getSystemService(Context.USB_SERVICE);
        List<UsbSerialDriver> availableDrivers = UsbSerialProber.getDefaultProber().findAllDrivers(manager);
        if (availableDrivers.isEmpty()) {
            logConsole("No UsbSerialDrivers");
            return;
        }
        else {
            logConsole("" + availableDrivers.size() + " drivers");
        }

        // Open a connection to the first available driver.
        UsbSerialDriver driver = availableDrivers.get(0);
        UsbDeviceConnection connection = manager.openDevice(driver.getDevice());
        if (connection == null) {
            logConsole("You probably need to call UsbManager.requestPermission(driver.getDevice()");
            return;
        }
        else
            logConsole("Connection opened");

        // Read some data! Most have just one port (port 0).
        List<UsbSerialPort> ports = driver.getPorts();
        if (ports.isEmpty()) {
            Log.d(TAG, "No ports");
            logConsole("No ports");
            return;
        }
        else
            logConsole(""+ports.size()+" ports");

        UsbSerialPort port = ports.get(0);
        sPort = port;

        /*try {
            port.open(connection);
            port.setParameters(9600, UsbSerialPort.DATABITS_8, UsbSerialPort.STOPBITS_1, UsbSerialPort.PARITY_NONE);
            byte buffer[] = new byte[16];
            int totalRead = 0;
            //while (totalRead < 100) {
                int numBytesRead = port.read(buffer, 1000);
                Log.d(TAG, "Read " + numBytesRead + " bytes.");
                logConsole("Read " + numBytesRead + " bytes.");
                totalRead += numBytesRead;
            //}
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            try {
                port.close();
            }
            catch (IOException e1) {
                e1.printStackTrace();
            }
        }*/

    }

    @Override
    protected void onPause() {
        super.onPause();

        mSensorManager.unregisterListener(this);

        stopIoManager();
        if (sPort != null) {
            try {
                sPort.close();
            } catch (IOException e) {
                // Ignore.
            }
            sPort = null;
        }
        finish();
    }


    @Override
    protected void onResume() {
        super.onResume();

        logConsole("Resumed, port=" + sPort);

        if (sPort == null) {
            //mTitleTextView.setText("No serial device.");
            logConsole("No serial device");
        } else {
            final UsbManager usbManager = (UsbManager) getSystemService(Context.USB_SERVICE);
            UsbDeviceConnection connection = usbManager.openDevice(sPort.getDriver().getDevice());
            if (connection == null) {
                logConsole("Opening device failed");
                return;
            }
            try {
                sPort.open(connection);
                sPort.setParameters(115200, UsbSerialPort.DATABITS_8, UsbSerialPort.STOPBITS_1, UsbSerialPort.PARITY_NONE);

                try {
                    //sendStatus();
                }
                catch (Exception e) {
                    StackTraceElement[] ste = e.getStackTrace();
                    for (int i = 0; i < ste.length; i++) {
                        logConsole( ste[i].toString() +"\n" );
                    }
                }

            } catch (IOException e) {
                logConsole("Error opening device: " + e.getMessage());
                try {
                    sPort.close();
                } catch (IOException e2) {
// Ignore.
                }
                sPort = null;
                return;
            }
            logConsole("Serial device: " + sPort.getClass().getSimpleName());
        }

        Sensor accelerometer = mSensorManager.getDefaultSensor(Sensor.TYPE_ACCELEROMETER);
        if (accelerometer != null) {
            mSensorManager.registerListener(this, accelerometer, SensorManager.SENSOR_DELAY_NORMAL, SensorManager.SENSOR_DELAY_UI);
        }

        Sensor magneticField = mSensorManager.getDefaultSensor(Sensor.TYPE_MAGNETIC_FIELD);
        if (magneticField != null) {
            mSensorManager.registerListener(this, magneticField, SensorManager.SENSOR_DELAY_NORMAL, SensorManager.SENSOR_DELAY_UI);
        }

        onDeviceStateChange();
    }
    private void stopIoManager() {
        if (mSerialIoManager != null) {
            logConsole("Stopping io manager ..\n");
            mSerialIoManager.stop();
            mSerialIoManager = null;
        }
    }
    private void startIoManager() {
        if (sPort != null) {
            logConsole("Starting io manager ..\n");
            mSerialIoManager = new SerialInputOutputManager(sPort, mListener);
            mExecutor.submit(mSerialIoManager);

            try {
                //sendStatus();
            }
            catch (Exception e) {
                StackTraceElement[] ste = e.getStackTrace();
                for (int i = 0; i < ste.length; i++) {
                    logConsole( ste[i].toString() +"\n" );
                }
            }
        }
    }
    private void onDeviceStateChange() {
        stopIoManager();
        startIoManager();
    }

    private Date data_rf24_time = null;
    private Date data_gsm_time = null;

    private int data_pps;
    private double data_age;
    private float data_lat;
    private float data_lon;
    private double data_hmsl;
    private double data_veld;
    private double data_gspeed;
    private double data_heading;

    private int data_sats;
    private int data_fix;
    private int data_csq;
    private int data_creg;

    private float data_gsm_lat;
    private float data_gsm_lon;
    private double data_gsm_hmsl;
    private double data_gsm_veld;
    private double data_gsm_gspeed;
    private double data_gsm_heading;

    private int data_gsm_sats;
    private int data_gsm_fix;
    private int data_gsm_csq;
    private double data_gsm_battery;

    private double data_ds1;
    private double data_ds2;
    private double data_MS5611_t;
    private double data_BMP280_t;
    private double data_Si7021_t;
    private double data_MPU6050_t;
    private double data_avginsidetemp_t;

    private double data_MS5611_p;
    private double data_BMP280_p;
    private double data_avgaltitude;

    private double data_Si7021_h;

    private double data_MPU6050_ax;
    private double data_MPU6050_ay;
    private double data_MPU6050_az;
    private double data_MPU6050_mag;

    private double data_battery;

    private double milliToMeter(int m) {
        double t = m / 100.0;
        t = Math.round(t);
        return t / 10.0;
    }

    private double hundredMilliToMeter(int m) {
        double t = m / 10000.0;
        t = Math.round(t);
        return t / 10.0;
    }

    private double mbarToMeters(double mbar) {
        double fraction = mbar / 1013.25;
        double raised = Math.pow( fraction, 0.190284 );
        return 145366.45 * ( 1 - raised ) * 0.3048;
    }

    private void parseData(String tmp) {

        try {
            JSONObject obj = new JSONObject(tmp);

            data_pps = obj.getInt("packets");
            data_age = milliToMeter(obj.getInt("dt"));

            data_lat = obj.getInt("lat") / 10000000.0f;
            data_lon = obj.getInt("lon") / 10000000.0f;
            data_hmsl = milliToMeter(obj.getInt("hmsl"));
            data_veld = milliToMeter(obj.getInt("veld"));
            data_gspeed = milliToMeter(obj.getInt("gspeed"));
            data_heading = hundredMilliToMeter(obj.getInt("heading"));

            data_sats = obj.getInt("sv");
            data_fix = obj.getInt("fix");
            data_csq = obj.getInt("csq");
            data_creg = obj.getInt("creg");

            data_ds1 = obj.getDouble("ds1");
            data_ds2 = obj.getDouble("ds2");

            data_MS5611_t = obj.getDouble("MS5611_t");
            data_BMP280_t = obj.getDouble("BMP280_t");
            data_Si7021_t = obj.getDouble("Si7021_t");
            data_MPU6050_t = obj.getDouble("MPU6050_t");
            double avginside = (data_MS5611_t + data_BMP280_t + data_Si7021_t + data_MPU6050_t) * 0.25;
            data_avginsidetemp_t = Math.round(avginside * 100.0) / 100.0;

            data_MS5611_p = obj.getDouble("MS5611_p");
            data_BMP280_p = obj.getDouble("BMP280_p");
            double agvPressure = (data_MS5611_p + data_BMP280_p) * 0.5;
            data_avgaltitude = Math.round(mbarToMeters(agvPressure) * 100.0) / 100.0;

            data_Si7021_h = obj.getDouble("Si7021_h");

            data_MPU6050_ax = obj.getDouble("MPU6050_ax");
            data_MPU6050_ay = obj.getDouble("MPU6050_ay");
            data_MPU6050_az = obj.getDouble("MPU6050_az");
            double mag = sqrt(data_MPU6050_ax * data_MPU6050_ax + data_MPU6050_ay * data_MPU6050_ay + data_MPU6050_az * data_MPU6050_az);
            data_MPU6050_mag = Math.round(mag * 100.0) / 100.0;
            data_battery = obj.getDouble("vbat");

            // only update latest time if pps indicates packets were actually received in this interval
            if ( data_pps > 0 ) {
                data_rf24_time = new Date();
            }

            lastNewDataTime = System.currentTimeMillis();
        }
        catch (JSONException e) {
            // ignore
            StackTraceElement[] ste = e.getStackTrace();
            for (int i = 0; i < ste.length; i++) {
                logConsole( ste[i].toString() );
            }
        }
        refreshData();
        refreshMap();
    }

    double lastBat = 0;

    private void refreshData() {

        float newestLat = 0;
        float newestLon = 0;
        double newestBat = 0;

        Date newestDate = null;
        if ( data_rf24_time != null && data_gsm_time != null ) {
            if ( data_gsm_time.compareTo(data_rf24_time) > 0 ) {
                newestDate = data_gsm_time;
                newestLat = data_gsm_lat;
                newestLon = data_gsm_lon;
                newestBat = data_gsm_battery;
            }
            else {
                newestDate = data_rf24_time;
                newestLat = data_lat;
                newestLon = data_lon;
                newestBat = data_battery;
            }
        }
        else if ( data_rf24_time != null ) {
            newestDate = data_rf24_time;
            newestLat = data_lat;
            newestLon = data_lon;
            newestBat = data_battery;
        }
        else if ( data_gsm_time != null ) {
            newestDate = data_gsm_time;
            newestLat = data_gsm_lat;
            newestLon = data_gsm_lon;
            newestBat = data_gsm_battery;
        }

        float uptake = 0.02f;
        newestBat = (uptake * newestBat) + ((1-uptake) * lastBat);

        lastBat = newestBat;

        String strDateFormat = "HH:mm:ss";
        DateFormat dateFormat = new SimpleDateFormat(strDateFormat);

        String formattedGSMDate =   data_gsm_time != null ? dateFormat.format(data_gsm_time) : "-";
        String formattedRF24Date =   data_rf24_time != null ? dateFormat.format(data_rf24_time) : "-";
        String formattedNewestDate = newestDate     != null ? dateFormat.format(newestDate) : "-";

        // this part doesn't work if not called from the main thread (eg. from GSM data checking task)
        //try {
            LatLng balloonPos = new LatLng(newestLat, newestLon);
            balloonMarker.setPosition(balloonPos);
        //}
        //catch (Exception e) {

        //}

        TabFragment1 tab1 = (TabFragment1) adapter.tab1;
        TabFragment2 tab2 = (TabFragment2) adapter.tab2;
        TabFragment3 tab3 = (TabFragment3) adapter.tab3;
        TabFragment4 tab4 = (TabFragment4) adapter.tab4;

        if (tab1 != null) {
            tab1.display_rf24_time.setText("Last time: " + formattedRF24Date);
            tab1.display_pps.setText("" + data_pps);
            tab1.display_age.setText("" + data_age);

            tab1.display_location.setText("" + data_lat + ", " + data_lon);
            tab1.display_hmsl.setText("" + data_hmsl + " m");
            tab1.display_veld.setText("" + data_veld + " m/s");
            tab1.display_gspeed.setText("" + data_gspeed + " m/s");
            tab1.display_heading.setText("" + data_heading + " deg");

            tab1.display_sats.setText("" + data_sats);
            tab1.display_fix.setText("" + data_fix);
            tab1.display_csq.setText("" + data_csq);
            tab1.display_creg.setText("" + data_creg);

            tab1.display_gsm_location.setText("" + data_gsm_lat + ", " + data_gsm_lon);
            tab1.display_gsm_hmsl.setText("" + data_gsm_hmsl + " m");
            tab1.display_gsm_veld.setText("" + data_gsm_veld + " m/s");
            tab1.display_gsm_gspeed.setText("" + data_gsm_gspeed + " m/s");
            tab1.display_gsm_heading.setText("" + data_gsm_heading + " deg");

            tab1.display_gsm_sats.setText("" + data_gsm_sats);
            tab1.display_gsm_fix.setText("" + data_gsm_fix);
            tab1.display_gsm_csq.setText("" + data_gsm_csq);

            tab1.display_gsm_time.setText("Last time: " + formattedGSMDate);
        }

        if (tab2 != null) {
            tab2.display_newestTime.setText("Latest time: " + formattedNewestDate);
            tab2.display_ds1.setText("" + data_ds2 + " °C");
            tab2.display_ds2.setText("" + data_ds1 + " °C");
            tab2.display_MS5611_t.setText("" + data_MS5611_t + " °C");
            tab2.display_BMP280_t.setText("" + data_BMP280_t + " °C");
            tab2.display_Si7021_t.setText("" + data_Si7021_t + " °C");
            tab2.display_MPU6050_t.setText("" + data_MPU6050_t + " °C");
            tab2.display_avginsidetemp.setText("" + data_avginsidetemp_t + " °C");

            tab2.display_MS5611_p.setText("" + data_MS5611_p + " mbar");
            tab2.display_BMP280_p.setText("" + data_BMP280_p + " mbar");
            tab2.display_avgaltitude.setText("" + data_avgaltitude + " m");

            tab2.display_Si7021_h.setText("" + data_Si7021_h + " %");

            tab2.display_MPU6050_xyz.setText("" + data_MPU6050_ax + ", " + data_MPU6050_ay + ", " + data_MPU6050_az);
            tab2.display_MPU6050_mag.setText("" + data_MPU6050_mag + " m/s²");

            String formattedBattery = String.format ("%,.2f", newestBat);
            tab2.display_battery.setText("Battery: " + formattedBattery + " V");
        }

        logToFile();
    }

    public void logToFile() {
        if ( fileOutputStream == null )
            return;

        // don't write more than once per five seconds
        if ( System.currentTimeMillis() < lastFileWriteTime + 5000 )
            return;

        // don't write if no new info has been received
        if ( lastFileWriteTime > lastNewDataTime )
            return;

        String formattedBattery = String.format ("%,.2f", lastBat);

        String strDateFormat = "yyyy/MM/dd HH:mm:ss";
        DateFormat dateFormat = new SimpleDateFormat(strDateFormat);

        String formattedRF24Date = "-";
        String formattedGSMDate = "-";
        if ( data_rf24_time != null ) {
            formattedRF24Date = dateFormat.format(data_rf24_time);
        }

        if ( data_gsm_time != null ) {
            formattedGSMDate = dateFormat.format(data_gsm_time);
        }

        try {
            fileOutputStream.write(("" + formattedRF24Date + ",").getBytes());
            fileOutputStream.write(("" + data_pps + ",").getBytes());
            fileOutputStream.write(("" + data_age + ",").getBytes());

            fileOutputStream.write(("" + data_lat + ",").getBytes());
            fileOutputStream.write(("" + data_lon + ",").getBytes());
            fileOutputStream.write(("" + data_hmsl + ",").getBytes());
            fileOutputStream.write(("" + data_veld + ",").getBytes());
            fileOutputStream.write(("" + data_gspeed + ",").getBytes());
            fileOutputStream.write(("" + data_heading + ",").getBytes());

            fileOutputStream.write(("" + data_sats + ",").getBytes());
            fileOutputStream.write(("" + data_fix + ",").getBytes());
            fileOutputStream.write(("" + data_csq + ",").getBytes());
            fileOutputStream.write(("" + data_creg + ",").getBytes());

            fileOutputStream.write(("" + formattedBattery + ",").getBytes());
            fileOutputStream.write(("" + data_ds1 + ",").getBytes());
            fileOutputStream.write(("" + data_ds2 + ",").getBytes());
            fileOutputStream.write(("" + data_MS5611_t + ",").getBytes());
            fileOutputStream.write(("" + data_BMP280_t + ",").getBytes());
            fileOutputStream.write(("" + data_Si7021_t + ",").getBytes());
            fileOutputStream.write(("" + data_MPU6050_t + ",").getBytes());
            fileOutputStream.write(("" + data_Si7021_h + ",").getBytes());
            fileOutputStream.write(("" + data_MPU6050_ax + ",").getBytes());
            fileOutputStream.write(("" + data_MPU6050_ay + ",").getBytes());
            fileOutputStream.write(("" + data_MPU6050_az + ",").getBytes());
            fileOutputStream.write(("" + data_MS5611_p + ",").getBytes());
            fileOutputStream.write(("" + data_BMP280_p + ",").getBytes());

            fileOutputStream.write(("" + formattedGSMDate + ",").getBytes());
            fileOutputStream.write(("" + data_gsm_lat + ",").getBytes());
            fileOutputStream.write(("" + data_gsm_lon + ",").getBytes());
            fileOutputStream.write(("" + data_gsm_hmsl + ",").getBytes());
            fileOutputStream.write(("" + data_gsm_veld + ",").getBytes());
            fileOutputStream.write(("" + data_gsm_gspeed + ",").getBytes());
            fileOutputStream.write(("" + data_gsm_heading + ",").getBytes());
            fileOutputStream.write(("" + data_gsm_sats + ",").getBytes());
            fileOutputStream.write(("" + data_gsm_fix + ",").getBytes());
            fileOutputStream.write(("" + data_gsm_csq + ",").getBytes());

            fileOutputStream.write(("\n").getBytes());

            fileOutputStream.flush();

        } catch (IOException e) {
            e.printStackTrace();
        }

        lastFileWriteTime = System.currentTimeMillis();
    }

    public void refreshMap() {
        if ( followType == 1 ) {
            map.moveCamera(CameraUpdateFactory.newLatLng( balloonMarker.getPosition() ));
        }
        else if ( followType == 2 ) {
            map.moveCamera(CameraUpdateFactory.newLatLng( selfMarker.getPosition() ));
        }
        else if ( followType == 3 ) {
            LatLngBounds.Builder builder = new LatLngBounds.Builder();
            builder.include(balloonMarker.getPosition());
            builder.include(selfMarker.getPosition());
            LatLngBounds bounds = builder.build();
            int padding = 100; // offset from edges of the map in pixels
            CameraUpdate cu = CameraUpdateFactory.newLatLngBounds(bounds, padding);
            map.moveCamera(cu);
        }
    }

    protected synchronized void buildGoogleApiClient() {
        GoogleApiClient.Builder builder = new GoogleApiClient.Builder(this);
        builder.addConnectionCallbacks(this);
        builder.addOnConnectionFailedListener(this);
        builder.addApi(LocationServices.API);
        mGoogleApiClient = builder.build();
        mGoogleApiClient.connect();
    }

    @Override
    public void onConnected(Bundle connectionHint) {
        Log.d(TAG, "onConnected");
        /*mLastLocation = LocationServices.FusedLocationApi.getLastLocation(
                mGoogleApiClient);
        if (mLastLocation != null) {
            //mLatitudeText.setText(String.valueOf(mLastLocation.getLatitude()));
            //mLongitudeText.setText(String.valueOf(mLastLocation.getLongitude()));
            LatLng loc = new LatLng(mLastLocation.getLatitude(), mLastLocation.getLongitude());
            CameraPosition cameraPosition = new CameraPosition.Builder().target(loc).zoom(18).build();
            mMap.animateCamera(CameraUpdateFactory.newCameraPosition(cameraPosition));
        }*/
        createLocationRequest();
    }

    protected void createLocationRequest() {
        LocationRequest mLocationRequest = new LocationRequest();
        mLocationRequest.setInterval(3000);
        mLocationRequest.setFastestInterval(1000);
        mLocationRequest.setSmallestDisplacement(1);
        mLocationRequest.setPriority(LocationRequest.PRIORITY_HIGH_ACCURACY);
        LocationServices.FusedLocationApi.requestLocationUpdates(mGoogleApiClient, mLocationRequest, this);
    }

    @Override
    public void onConnectionFailed(ConnectionResult connectionResult) {
        Log.d(TAG, "onConnectionFailed");
    }

    @Override
    public void onConnectionSuspended(int i) {
        Log.d(TAG, "onConnectionSuspended");
    }

    public void onLocationChanged(Location location) {
        Log.d(TAG, "onLocationChanged "+location.toString());

        double x = location.getLatitude();
        double y = location.getLongitude();
        LatLng loc = new LatLng(x,y);
        selfMarker.setPosition(loc);

        if ( followType > 0 )
            refreshMap();

        selfAltitude = location.getAltitude();
    }

    private static final CharSequence[] MAP_TYPE_ITEMS = {"Road Map", "Satellite", "Terrain", "Hybrid"};

    private void showMapTypeSelectorDialog() {
        // Prepare the dialog by setting up a Builder.
        final String fDialogTitle = "Select map type";
        AlertDialog.Builder builder = new AlertDialog.Builder(this);
        builder.setTitle(fDialogTitle);

        // Find the current map type to pre-check the item representing the current state.
        int checkItem = map.getMapType() - 1;

        // Add an OnClickListener to the dialog, so that the selection will be handled.
        builder.setSingleChoiceItems(
                MAP_TYPE_ITEMS,
                checkItem,
                new DialogInterface.OnClickListener() {
                    public void onClick(DialogInterface dialog, int item) {
                        map.setMapType(item + 1);
                        dialog.dismiss();
                    }
                }
        );

        // Build the dialog and show it.
        AlertDialog fMapTypeDialog = builder.create();
        fMapTypeDialog.setCanceledOnTouchOutside(true);
        fMapTypeDialog.show();
    }

    private int followType = 0;
    private static final CharSequence[] FOLLOW_TYPE_ITEMS = { "Nothing", "Balloon", "Tablet", "Both"};

    private void showFollowTypeSelectorDialog() {
        final String fDialogTitle = "Select follow type";
        AlertDialog.Builder builder = new AlertDialog.Builder(this);
        builder.setTitle(fDialogTitle);

        // Add an OnClickListener to the dialog, so that the selection will be handled.
        builder.setSingleChoiceItems(
                FOLLOW_TYPE_ITEMS,
                followType,
                new DialogInterface.OnClickListener() {
                    public void onClick(DialogInterface dialog, int item) {
                        followType = item;
                        refreshMap();
                        dialog.dismiss();
                    }
                }
        );

        // Build the dialog and show it.
        AlertDialog fMapTypeDialog = builder.create();
        fMapTypeDialog.setCanceledOnTouchOutside(true);
        fMapTypeDialog.show();
    }

    // Get readings from accelerometer and magnetometer. To simplify calculations,
    // consider storing these readings as unit vectors.
    @Override
    public void onSensorChanged(SensorEvent event) {
        if (event.sensor.getType() == Sensor.TYPE_ACCELEROMETER) {
            System.arraycopy(event.values, 0, mAccelerometerReading,0, mAccelerometerReading.length);
            updateOrientationAngles();
        } else if (event.sensor.getType() == Sensor.TYPE_MAGNETIC_FIELD) {
            System.arraycopy(event.values, 0, mMagnetometerReading,0, mMagnetometerReading.length);
            updateOrientationAngles();
        }
    }

    @Override
    public void onAccuracyChanged(Sensor sensor, int accuracy) {
        // Do something here if sensor accuracy changes.
        // You must implement this callback in your code.
    }

    // Compute the three orientation angles based on the most recent readings from
    // the device's accelerometer and magnetometer.
    public void updateOrientationAngles() {
        // Update rotation matrix, which is needed to update orientation angles.
        SensorManager.getRotationMatrix(mRotationMatrix, null, mAccelerometerReading, mMagnetometerReading);

        // "mRotationMatrix" now has up-to-date information.

        SensorManager.getOrientation(mRotationMatrix, mOrientationAngles);

        // "mOrientationAngles" now has up-to-date information.
        //Log.d(TAG, ""+mOrientationAngles[0] / Math.PI * 180);
        refreshCompass();
    }

    private void refreshCompass() {
        TabFragment3 tab3 = (TabFragment3) adapter.tab3;
        if ( tab3 != null ) {
            float tmp = mOrientationAngles[0] * 180.f / (float)Math.PI;
            float declination = 20;
            tmp += declination;

            float delta = tmp - compassDirection;
            while ( delta > 180 )
                delta += -360;
            while ( delta < -180 )
                delta += 360;

            compassDirection += delta * 0.2f;

            while (compassDirection > 360) compassDirection -= 360;
            while (compassDirection <   0) compassDirection += 360;

            double bearingToTarget = geoBearing(selfMarker.getPosition(), balloonMarker.getPosition());
            double distanceToTarget = geoDistance(selfMarker.getPosition(), balloonMarker.getPosition());

            while (bearingToTarget > 360) bearingToTarget -= 360;
            while (bearingToTarget <   0) bearingToTarget += 360;

            tab3.display_compass.setText( ""+((int)Math.round(compassDirection))+" degrees" );
            tab3.display_bearing.setText( ""+((int)Math.round(bearingToTarget))+" degrees" );


            String distanceUnits = "m";
            if ( distanceToTarget > 10000 ) {
                distanceToTarget *= 0.001;
                distanceUnits = "km";
            }
            String formattedDistance = String.format ("%,.0f", distanceToTarget);
            tab3.display_distance.setText( "Distance: "+formattedDistance+" "+distanceUnits );


            double relativeHeight = data_hmsl - selfAltitude;
            String heightUnits = "m";
            if ( relativeHeight > 10000 ) {
                relativeHeight *= 0.001;
                heightUnits = "km";
            }
            String formattedRelativeHeight = String.format ("%,.0f", relativeHeight);
            tab3.display_relativeHeight.setText( "Height: "+formattedRelativeHeight+" "+heightUnits );


            tab3.display_arrowImage.setRotation( (float)bearingToTarget - compassDirection );
        }
    }

    public final double DEGTORAD = 0.0174532925199432957;
    public final double RADTODEG = 57.295779513082320876;

    // http://www.movable-type.co.uk/scripts/latlong.html
    double geoDistance(LatLng a, LatLng b) {
        double R = 6371000; // km
        double p1 = a.latitude * DEGTORAD;
        double p2 = b.latitude * DEGTORAD;
        double dp = (b.latitude - a.latitude) * DEGTORAD;
        double dl = (b.longitude - a.longitude) * DEGTORAD;

        double x = sin(dp/2) * sin(dp/2) + cos(p1) * cos(p2) * sin(dl/2) * sin(dl/2);
        double y = 2 * atan2(sqrt(x), sqrt(1-x));

        return R * y;
    }

    double geoBearing(LatLng a, LatLng b) {
        double y = sin(b.longitude - a.longitude) * cos(b.latitude);
        double x = cos(a.latitude)*sin(b.latitude) - sin(a.latitude)*cos(b.latitude)*cos(b.longitude - a.longitude);
        return atan2(y, x) * RADTODEG;
    }





    private class CheckServerReports extends AsyncTask<String, Integer, Long> {
        protected Long doInBackground(String... params) {
            int a = 2;
            while (a == 2) {

                try {
                    Thread.sleep(4000);
                }
                catch (InterruptedException e) {
                    Thread.currentThread().interrupt();
                }

                try {
                    URL url = new URL(params[0]);
                    HttpURLConnection connection = (HttpURLConnection) url.openConnection();
                    connection.connect();

                    InputStream stream = connection.getInputStream();

                    BufferedReader reader = new BufferedReader(new InputStreamReader(stream));

                    StringBuffer buffer = new StringBuffer();
                    String line = "";

                    while ((line = reader.readLine()) != null) {
                        buffer.append(line+"\n");
                        //Log.d("Response: ", "> " + line);
                    }

                    try {

                        JSONObject obj = new JSONObject(String.valueOf(buffer));

                        int id = obj.getInt("id");

                        Log.d(TAG, "id "+id);

                        if ( id > highestGSMReportId ) {

                            Log.d(TAG, "Doing GSM update ");

                            String timeStr = obj.getString("jst");
                            String latlon = obj.getString("latlon");
                            int battery = obj.getInt("battery");
                            data_gsm_csq = obj.getInt("signal_quality");
                            data_gsm_fix = obj.getInt("fix");
                            data_gsm_sats = obj.getInt("sats");
                            data_gsm_heading = hundredMilliToMeter(obj.getInt("heading"));
                            data_gsm_gspeed = milliToMeter(obj.getInt("speed"));
                            data_gsm_veld = milliToMeter(obj.getInt("veld"));
                            data_gsm_hmsl = milliToMeter(obj.getInt("height"));
                            String MPU6050 = obj.getString("MPU6050");
                            String MS5611 = obj.getString("MS5611");
                            String BMP280 = obj.getString("BMP280");
                            String Si7021 = obj.getString("Si7021");
                            String DS18B20 = obj.getString("DS18B20");

                            String[] latlonArray = latlon.split(",");
                            data_gsm_lat = Float.valueOf(latlonArray[0]) / 10000000.0f;
                            data_gsm_lon = Float.valueOf(latlonArray[1]) / 10000000.0f;

                            SimpleDateFormat myFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
                            try {
                                data_gsm_time = myFormat.parse(timeStr);

                                Calendar cal = Calendar.getInstance();
                                cal.setTime(data_gsm_time);
                                cal.add(Calendar.HOUR_OF_DAY, 4);
                                data_gsm_time = cal.getTime();
                            } catch (ParseException e) {
                                e.printStackTrace();
                            }

                            float r1 = 3280;
                            float r2 = 9991;

                            data_gsm_battery = ((battery / 1023.0f) * 3.3) * ((r1 + r2) / r1);

                            //Log.d(TAG, "Got GSM data, battery = " + data_gsm_battery);

                            highestGSMReportId = id;
                            lastNewDataTime = System.currentTimeMillis();

                            runOnUiThread(new Runnable() {
                                @Override
                                public void run() {
                                    refreshData();
                                }
                            });
                        }
                    }
                    catch (Exception je) {
                        Log.d(TAG, je.toString());
                    }

                }
                catch (Exception e) {
                    Log.d(TAG, e.toString());
                }
            }
            return 0L;
        }

        protected void onProgressUpdate(Integer... progress) {
            //Log.d(TAG, ""+progress[0]);
        }

        protected void onPostExecute(Long result) {
            Log.d(TAG, "thread done");
        }
    }
}
